-- Monster Arena Reloaded
-- Nox script by KirConjurer

dofile(".\\scripts\\unitUtil.lua")

-- Team spawn coordinates
MobArena.SpawnLocations = { { 1943, 1805 }, { 4128, 4174 } }
-- Structure coordinates
MobArena.BaseLocations = { { 2530, 1633 }, { 3520, 4370 } }
-- Experimental stuff
MobArena.AllowExperimental = false
MobArena.UseRespawnSystem = false

MobArena.PlayerToMonsterMap = {}
MobArena.MonsterToPlayerMap = {}
MobArena.MonsterToSpawnClass = {}
MobArena.DisallowMorphUnits = {}
-- [monster][effid] = duration
MobArena.CustomEffects = {}
local CustomEffectID = {
    NATURES_GIFT = 1,
    MORPH_FROG = 2,
    MORPH_BAT = 3,
    CHANNELING = 4,
    WEAKNESS = 5
}

if (MobArena.AllowExperimental) then
    monsterRegisterFireSprite()
    --[[
    MobArena.fspriteSoundSet = memAlloc(0x54)
    setPtrInt(MobArena.fspriteSoundSet, 0x18, 0x36F) -- MELEE_ATTACK_SOUND
    setPtrInt(MobArena.fspriteSoundSet, 0x38, 9) -- DIE_SOUND
    setPtrInt(MobArena.fspriteSoundSet, 0x3C, 0x11B) -- INJURED_SOUND
    ]]--
end

MobArena.Refresh = function()
    if hostPlayer then
        if MobArena.PlayerToMonsterMap[hostPlayer] then
            local monster = MobArena.PlayerToMonsterMap[hostPlayer]
            local class = MobArena.MonsterToSpawnClass[monster]
            class.AbilityLastUseTime[1] = 0
            class.AbilityLastUseTime[2] = 0
            class.AbilityLastUseTime[3] = 0
        end
    end
end

local function utilGetSpawnPos(team)
    if (team < 1 or team > 2) then
        print("Wrong team id passed to utilGetSpawnPos: " .. team)
    else
        return MobArena.SpawnLocations[team]
    end
    return nil
end

local function utilAtPlayerMouse(plr)
    local x, y = playerMouse(plr)
    local result = nil
    local currDist = 999
    local tempDist
    local unitRadius
    local fn = function(p) 
        if (not unitIsDead(p) and unitIsMonster(p)) then 
            tempDist = distance(x, y, unitPos(p))
            unitRadius = getPtrFloat(p, 0xB0) + 2
            if tempDist <= unitRadius and currDist > tempDist then
                currDist = tempDist
                result = p
            end
        end
    end
    local search = 20
    unitGetAround(fn, x - search, y - search, x + search, y + search)
    return result
end

local function utilAbilityCDString(abil, usetime)
    local frames = usetime - getFrameCounter()
    local sec = math.floor(frames / utilGetGameFPS()) + 1
    local result
    if sec == 1 then
        result = string.format("%s: 1 second", abil)
    else
        result = string.format("%s: %i seconds", abil, sec)
    end
    return result
end

function utilSetActionTargeted(mob, act, x, y)
    local t = unitSetAction(mob, act)
    setPtrFloat(t, 4, x)
    setPtrFloat(t, 8, y)
    setPtrInt(t, 12, 0)
    return t
end

function utilSetActionFight(mob, targ)
    local t = unitSetAction(mob, 15) -- ACTION_FIGHT
    local tx, ty = unitPos(targ)
    setPtrFloat(t, 4, tx)
    setPtrFloat(t, 8, ty)
    setPtrInt(t, 12, getFrameCounter())
    return t
end

local function utilSetActionCastDuration(mob, id)
    local t = unitSetAction(mob, 20) -- ACTION_CAST_DURATION_SPELL
    setPtrInt(t, 4, id)
    setPtrInt(t, 8, 0)
    setPtrInt(t, 12, 0)
    return t
end

local function utilSetActionCastTo(mob, x, y, id)
    local t = unitSetAction(mob, 19) -- ACTION_CAST_SPELL_ON_LOCATION
    setPtrInt(t, 4, id)
    setPtrInt(t, 8, 0)
    setPtrFloat(t, 12, x)
    setPtrFloat(t, 16, y)
    setPtrFloat(t, 20, 0)
    return t
end

function distance(x1, y1, x2, y2)
    local dx = x1 - x2
    local dy = y1 - y2
    return math.sqrt(dx * dx + dy * dy)
end

local function monsterDefaultMeleeAttack(mob, x, y)
    -- push action
    unitActionPop(mob)
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    unitSetAction(mob, 16) -- ACTION_MELEE_ATTACK
end

local function monsterDefaultMissileAttack(mob, x, y)
    unitActionPop(mob)
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    utilSetActionTargeted(mob, 17, x, y) -- ACTION_MISSILE_ATTACK
end

local function monsterWaitForRespawn(mob)
    local plr = MobArena.MonsterToPlayerMap[mob]
    if objectExists(plr) then
        playerOnInput[plr] = function() return true end
        local dummy = createObject("ImaginaryCaster", unitPos(mob))
        playerLook(plr, dummy)
        MobArena.PlayerToMonsterMap[plr] = nil
        MobArena.MonsterToPlayerMap[mob] = nil
        local class = MobArena.MonsterToSpawnClass[mob]
        MobArena.MonsterToSpawnClass[mob] = nil
        setTimeout(function()
            unitDelete(dummy)
            -- make sure player didn't respawned 
            if (objectExists(plr) and MobArena.PlayerToMonsterMap[plr] == nil) then
                class.Respawn(unitGetTeam(plr))
                class.SetPlayer(plr)
                class.AbilityLastUseTime[1] = 0
                class.AbilityLastUseTime[2] = 0
                class.AbilityLastUseTime[3] = 0
                MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
            end
        end, utilSecToFrames(5, 1))
    end
end

local function monsterWaitNoRespawn(mob)
    local plr = MobArena.MonsterToPlayerMap[mob]
    if objectExists(plr) then
        playerOnInput[plr] = nil
        local dummy = createObject("ImaginaryCaster", unitPos(mob))
        playerLook(plr, dummy)
        MobArena.PlayerToMonsterMap[plr] = nil
        MobArena.MonsterToPlayerMap[mob] = nil
        local class = MobArena.MonsterToSpawnClass[mob]
        MobArena.MonsterToSpawnClass[mob] = nil
        setTimeout(function()
            if objectExists(plr) then
                if ptrCall2(0x4DDF30, plr, plr) == dummy then
                    playerLook(plr, nil)
                end
            end
            unitDelete(dummy)
        end, utilSecToFrames(5, 1))
    end
end

function utilIsMagicVulnerable(targ)
    if (unitTestBuff(targ, 23) or unitTestBuff(targ, 19)) then return false end
    return true -- check ENCHANT_INVULNERABLE and ENCHANT_PROTECT_FROM_MAGIC (unused by Westwood, but now it's magic immunity)
end

function monsterResetAttackDelay(mob)
    setPtrInt(getPtrPtr(mob, 0x2EC), 0x200, 0)
end

local function utilRemovePlayerHandlers(plr)
    local mob = MobArena.PlayerToMonsterMap[plr]
    playerOnInput[plr] = nil
    playerLook(plr, nil)
    if (mob) then
        MobArena.PlayerToMonsterMap[plr] = nil
        MobArena.MonsterToPlayerMap[mob] = nil
        if objectExists(mob) then unitDelete(mob)end
    end
end

local function utilFindClosestHostileMonster(x, y, rel, dist)
    local result = nil
    local closest = dist * dist
    local px, py
    local rx, ry = unitPos(rel)
    local fn = function(p)
        if (unitIsMonster(p) and unitCanAttack(rel, p)) then
            px, py = unitPos(p)
            if (distance(px, py, x, y) < closest and mapTraceRay(px, py, rx, ry, 5)) then
                closest = distance(px, py, x, y)
                result = p
            end
        end 
    end
    unitGetAround(fn, x - dist, y - dist, x + dist, y + dist)
    return result
end

local function utilFindClosestFriendlyMonster(x, y, rel, dist)
    local result = nil
    local closest = dist * dist
    local px, py
    local rx, ry = unitPos(rel)
    local fn = function(p)
        if (unitIsMonster(p) and not unitCanAttack(rel, p)) then
            px, py = unitPos(p)
            if (distance(px, py, x, y) < closest and mapTraceRay(px, py, rx, ry, 5)) then
                closest = distance(px, py, x, y)
                result = p
            end
        end 
    end
    unitGetAround(fn, x - dist, y - dist, x + dist, y + dist)
    return result
end

local function utilCheckCustomEffect(mob, id)
    if (MobArena.CustomEffects[mob]) then
        if (MobArena.CustomEffects[mob][id]) then return true end
    end
    return false
end

local function utilMonsterCanAttack(mob)
    -- no ENCHANT_FREEZE or ENCHANT_CONFUSED or ENCHANT_AFRAID or ENCHANT_ANTI_MAGIC
    if unitTestBuff(mob, 25) or unitTestBuff(mob, 3) or unitTestBuff(mob, 11) or unitTestBuff(mob, 29) then return false end
    if utilCheckCustomEffect(mob, CustomEffectID.CHANNELING) then return false end
    return true
end

local function utilApplyCustomEffect(mob, id, duration)
    if not MobArena.CustomEffects[mob] then
        MobArena.CustomEffects[mob] = {}
    end
    MobArena.CustomEffects[mob][id] = getFrameCounter() + duration
    
    -- Morph effect needs special handling
    if (id >= CustomEffectID.MORPH_FROG and id <= CustomEffectID.MORPH_BAT) then
        local monsterName = "GreenFrog"
        if (id == CustomEffectID.MORPH_BAT) then monsterName = "Bat" end
        
        if MobArena.DisallowMorphUnits[mob] then return end
        MobArena.DisallowMorphUnits[mob] = true -- fix
        local replace = createObject(monsterName, unitPos(mob))
        setPtrInt(mob, 0x10, bitOr(getPtrInt(mob, 0x10), 2))
        unitMove(mob, 0, 0) -- move original monster out of the map
        if (id == CustomEffectID.MORPH_BAT) then
            unitSetMaxHealth(replace, 0) -- make bat invulnerable
            setPtrFloat(replace, 0x224, getPtrFloat(replace, 0x224) * 0.7) -- reduce speed
        else
            unitSetMaxHealth(replace, unitGetMaxHP(mob))
            unitSetCurrentHP(replace, unitGetHPSafe(mob))
            setPtrFloat(replace, 0x224, getPtrFloat(replace, 0x224) * 0.4) -- reduce speed
        end
        unitSetLookAngle(replace, unitGetLookAngle(mob))
        
        local plr = MobArena.MonsterToPlayerMap[mob]
        if (plr) then
            unitBecomePet(replace, plr)
            playerLook(plr, replace)
            unitAddRecentlyCreated()
            unitSetAgression(replace, 0)
            local uc = getPtrPtr(replace, 0x2EC)
            setPtrFloat(uc, 0x54C, 0) -- no AI
            setPtrFloat(uc, 0x538, 0) -- no AI
            setPtrInt(uc, 0x5A0, 0) -- reset monster status
            
            local moveTime = 0
            playerOnInput[plr] = function(plr2, c)
                if c == 1 then
                    if moveTime < getFrameCounter() and monsterGetCurrentAction(replace) == 9 then 
                        unitActionPop(replace)
                        moveTime = getFrameCounter() + 1
                    end
                elseif c == 2 then 
                    if moveTime < getFrameCounter() then
                        local mx, my = unitPos(replace)
                        local x, y = playerMouse(plr)
                        moveTime = getFrameCounter() + 1
                        if (x > 0 and y > 0 and x < 5887 and y < 5887 and distance(x, y, mx, my) > 5) then
                            unitSetLookAngle(replace, directGet(x, y, mx, my))
                            unitActionPop(replace)
                            utilSetActionTargeted(replace, 9, x, y) -- ACTION_DODGE
                        end
                    end
                end
                return true
            end
        end
        local isDead = false
        local restoreOriginal = function()
            if not isDead then
                if (id ~= CustomEffectID.MORPH_BAT) then
                    unitSetCurrentHP(mob, unitGetHPSafe(replace))
                end
                MobArena.DisallowMorphUnits[mob] = nil
                unitMove(mob, unitPos(replace)) -- swap with original monster
                unitSetLookAngle(mob, unitGetLookAngle(replace))
                setPtrInt(mob, 0x10, bitXor(getPtrInt(mob, 0x10), 2))
                if (plr) then
                    playerLook(plr, mob)
                    playerOnInput[plr] = MobArena.MonsterToSpawnClass[mob].PlayerInput
                end
                netPointFx(0x8A, unitPos(replace))
                unitDelete(replace)
            end
        end
        unitOnDie(replace, function()
            restoreOriginal()
            unitDamagePure(mob, 9999)
            isDead = true
        end)
        setTimeout(restoreOriginal, duration)
    end
end

playerOnGoObs = utilRemovePlayerHandlers
playerOnLeave = utilRemovePlayerHandlers
_G.unitOnCastList = {} -- bugfix
_G.unitFlyActivate = {}

-- create generic monster class
function NewMonsterClass()
    local self = 
    {
        AbilityKeyHandlers = { nil, nil, nil },
        AbilityUseTimeout = { -1, -1, -1 },
        AbilityLastUseTime = { 0, 0, 0 },
        AbilityNames = { "", "", "" },
        AttackHandler = nil,
        AttackTimeout = -1,
        LastAttackTime = 0,
        ThingID = "",
        MonsterPtr = nil,
        BaseHealth = -1,
        MoveSpeedMul = 0.8
    }
    
    -- set by unitOnDie
    local monsterIsDead = false
    -- player ID 0-31
    local plid = -1
    
    -- respawn monster with full health
    function self.Respawn(team)
        local spawnPos = utilGetSpawnPos(team)
        
        -- Spawn monster
        self.MonsterPtr = createObject(self.ThingID, spawnPos[1], spawnPos[2])
        unitSetMaxHealth(self.MonsterPtr, self.BaseHealth) -- set health
        unitSetAgression(self.MonsterPtr, 0) -- no AI
        monsterNoSpells(self.MonsterPtr)
        setPtrFloat(self.MonsterPtr, 0x224, getPtrFloat(self.MonsterPtr, 0x224) * self.MoveSpeedMul) -- multiply speed
        local uc = getPtrPtr(self.MonsterPtr, 0x2EC)
        setPtrFloat(uc, 0x54C, 0) -- no AI
        setPtrFloat(uc, 0x538, 0) -- no AI
        setPtrInt(uc, 0x5A0, 0) -- reset monster status
        buffApply(self.MonsterPtr, 23, utilSecToFrames(4, 1)) -- ENCHANT_INVULNERABLE
        unitNoPoisonResist(self.MonsterPtr) -- remove poison immunity
    end
    
    -- register handlers and stuff
    function self.SetPlayer(parentPlr)
        if not self.MonsterPtr then return end
        unitBecomePet(self.MonsterPtr, parentPlr) -- set parent
        MobArena.PlayerToMonsterMap[parentPlr] = self.MonsterPtr
        MobArena.MonsterToPlayerMap[self.MonsterPtr] = parentPlr
        playerOnInput[parentPlr] = self.PlayerInput
        plid = playerInfo(parentPlr).idx
        unitOnDie(self.MonsterPtr, function(ignore, attacker)
            if (self.MonsterDie) then self.MonsterDie(self.MonsterPtr, attacker) end
            unitDecay(self.MonsterPtr, 15)
            if (MobArena.UseRespawnSystem) then
                monsterWaitForRespawn(self.MonsterPtr)
                sendChat(plid, "Monster has been slain! Respawning in 5 seconds...", servNetCode(self.MonsterPtr), false)
            else
                monsterWaitNoRespawn(self.MonsterPtr)
            end
        end)
        -- TODO unitOnDie, unitOnDamage
        unitAddRecentlyCreated()
        playerLook(parentPlr, self.MonsterPtr)
    end
    
    local moveTime = 0
    local callTime = 0
    
    -- playerOnInput handler
    function self.PlayerInput(pl, c, targ)
        local x, y = playerMouse(pl)
        local fram = getFrameCounter()
        local mob = self.MonsterPtr
  
        if callTime < fram then
            -- useful for chanelling abilities
            if (self.EveryFrame) then self.EveryFrame() end
            callTime = fram
            -- apply custom effects
            effects = MobArena.CustomEffects[mob]
            if (effects) then
                local dur1 = effects[CustomEffectID.NATURES_GIFT]
                if (dur1) then
                    if (dur1 < fram) then 
                        -- duration is over - remove effect
                        effects[CustomEffectID.NATURES_GIFT] = nil
                        soundMake(124, unitPos(mob))
                    else
                        -- spawn particles
                        if (math.mod(fram, 2) == 0) then
                            mx, my = unitPos(mob)
                            mx = mx + math.random(-20, 20)
                            my = my + 15
                            local orb = createObject("CharmOrb", mx, my)
                            netLinearOrbMove(orb, 16, 8, 0.07, 7)
                        end
                        -- regeneration
                        if (math.mod(fram, 8) == 0) then unitAddHP(mob, 1) end
                    end
                end
                dur1 = effects[CustomEffectID.CHANNELING]
                if (dur1) then if (dur1 < fram) then effects[CustomEffectID.CHANNELING] = nil end end
            end
        end
  
        -- if monster is dead or channeling active then return immediately
        if unitIsDead(mob) or utilCheckCustomEffect(mob, CustomEffectID.CHANNELING) then return true end
        
        if c == 1 then -- no player action
            -- look around
            if (monsterGetCurrentAction(mob) == 0) then
                unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
            end
            -- stop continous movement
            if moveTime and moveTime < fram and monsterGetCurrentAction(mob) == 9 then 
                unitActionPop(mob)
                moveTime = fram + 1
            end
        elseif c == 2 then -- move to 
            if (not moveTime or moveTime < fram) and monsterGetCurrentAction(mob) ~= 19 then
                moveTime = fram + 10
                local mx, my = unitPos(mob)
                if (x > 0 and y > 0 and x < 5887 and y < 5887 and distance(x, y, mx, my) > 5) then
                    unitSetLookAngle(mob, directGet(x, y, mx, my))
                    unitActionPop(mob)
                    utilSetActionTargeted(mob, 9, x, y) -- ACTION_DODGE
            
                    -- play audio effect
                    ptrCall2(0x534030, mob, mob) -- just use original function that does this
                    --[[
                    local ss = monsterGetSoundset(mob)
                    local moveSound = getPtrInt(ss, 72)
                    local animFrame = getPtrByte(getPtrPtr(mob, 0x2EC), 0x1E1)
                    if moveSound and (getPtrInt(ss, 0x64) == animFrame or getPtrInt(ss, 0x68) == animFrame) then soundMake(moveSound, mx, my) end 
                    ]]--
                end
            end
        elseif c == 7 then
            if (self.SpacePressed) then self.SpacePressed() end
        end
        
        if utilMonsterCanAttack(mob) then
            if c == 6 then -- attack
                if (self.LastAttackTime < fram) and monsterGetCurrentAction(mob) ~= 19 then
                    monsterResetAttackDelay(mob)
                    self.LastAttackTime = fram + self.AttackTimeout
                    if (self.AttackHandler) then self.AttackHandler(mob, x, y) end
                    -- Nature's Gift - increase attack speed
                    if (utilCheckCustomEffect(mob, CustomEffectID.NATURES_GIFT) and self.AttackTimeout >= 15) then
                        self.LastAttackTime = self.LastAttackTime - 5
                    end
                end
            elseif c == 47 then -- use first ability
                if (self.AbilityLastUseTime[1] < fram) then
                    if (self.AbilityKeyHandlers[1]) then
                        if not self.AbilityKeyHandlers[1](mob, x, y) then
                            self.AbilityLastUseTime[1] = fram + self.AbilityUseTimeout[1]
                        end
                    end
                else
                    sendChat(plid, utilAbilityCDString(self.AbilityNames[1], self.AbilityLastUseTime[1]), servNetCode(mob), false)
                end
            elseif c == 49 then -- use second ability
                if (self.AbilityLastUseTime[2] < fram) then
                    if (self.AbilityKeyHandlers[2]) then
                        if not self.AbilityKeyHandlers[2](mob, x, y) then
                            self.AbilityLastUseTime[2] = fram + self.AbilityUseTimeout[2]
                        end
                    end
                else
                    sendChat(plid, utilAbilityCDString(self.AbilityNames[2], self.AbilityLastUseTime[2]), servNetCode(mob), false)
                end
            elseif c == 48 then -- use third ability
                if (self.AbilityLastUseTime[3] < fram) then
                    if (self.AbilityKeyHandlers[3]) then
                        if not self.AbilityKeyHandlers[3](mob, x, y) then
                            self.AbilityLastUseTime[3] = fram + self.AbilityUseTimeout[3]  
                        end
                    end
                else
                    sendChat(plid, utilAbilityCDString(self.AbilityNames[3], self.AbilityLastUseTime[3]), servNetCode(mob), false)
                end
            end
        end
        return true
    end
    
    return self
end

function MobArena.spawnSkeleton(plr)
    -- ability related constants
    local ability1Duration = 180
    local ability1VampPower = 5
    local ability2Duration = 90

    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "SkeletonLord"
    class.BaseHealth = 60
    class.AttackTimeout = 20
    class.AttackHandler = monsterDefaultMeleeAttack

    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(14, 1)
    class.AbilityNames[1] = "Undead Power"
    class.AbilityKeyHandlers[1] = function()
        -- play audio
        soundMake(912, unitPos(class.MonsterPtr))
        -- apply vampirism enchant
        buffApplySilent(class.MonsterPtr, 13, ability1Duration, ability1VampPower)
        -- increase attack speed
        local oldAT = class.AttackTimeout
        class.AttackTimeout = 9
        setTimeout(function() class.AttackTimeout = oldAT end, ability1Duration)
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(10, 1)
    class.AbilityNames[2] = "Bat Form"
    class.AbilityKeyHandlers[2] = function(mob)
        soundMake(338, unitPos(mob))
        utilApplyCustomEffect(mob, CustomEffectID.MORPH_BAT, ability2Duration)
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnBear(plr)
    -- ability related constants
    local meleeAttackFrame = 8
    local forceStunDuration = 15
    local forceBonusDamage = 9
    local hasteDuration = 70
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "BlackBear"
    class.BaseHealth = 110
    class.AttackTimeout = 45
    class.AttackHandler = monsterDefaultMeleeAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(8, 1)
    class.AbilityNames[1] = "Shockforce"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) monsterDefaultMeleeAttack(mob, x, y)
        -- wait for attack to finish
        setTimeout(function()
            if unitIsDead(mob) then return end
            if monsterGetCurrentAction(mob) == 16 then -- ACTION_MELEE_ATTACK
                local target = monsterPickMeleeStrike(mob)
                if not target then return end
                if not unitCanAttack(mob, target) then return end
                buffApply(target, 5, forceStunDuration)
                unitDamageHack(target, NOX_DAMAGE_TYPE.CLAW, forceBonusDamage, mob)
                local tx, ty = unitPos(target)
                netSendEarthquake(tx, ty, 10)
            end
        end, meleeAttackFrame)
    end
    
    class.EveryFrame = function()
        -- bonus health regeneration
        if math.mod(getFrameCounter(), 8) == 0 and unitLastDamageDiff(class.MonsterPtr) > 60 then
            unitAddHP(class.MonsterPtr, 1)
        end
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnMechFlyer(plr)
    -- ability related constants

    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "FlyingGolem"
    class.BaseHealth = 50
    class.AttackTimeout = 12
    class.AttackHandler = monsterDefaultMissileAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(4, 1)
    class.AbilityNames[1] = "Triple Shot"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) -- stop movement
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        local look = unitGetLookAngle(mob)
        local shot, vx, vy
        for ia = -7, 7, 7 do -- spawn 3 arrows
            vx = utilCosFromDirect(look + ia) * 50
            vy = utilSinFromDirect(look + ia) * 50
            shot = createObject("GolemArrow", unitPos(mob))
            setPtrFloat(shot, 0x50, vx)
            setPtrFloat(shot, 0x54, vy)
            unitSetLookAngle(shot, look)
            unitSetParent(shot, mob) -- set parent
            arrowSetDamage(shot, 10) -- increase damage
        end
        unitAddRecentlyCreated()
        soundMake(889, unitPos(mob))
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(16, 1)
    class.AbilityNames[2] = "Self Repair"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        local mx, my = unitPos(mob)
        soundMake(811, mx, my)
        soundMake(803, mx, my)
        unitAddHP(mob, class.BaseHealth)
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnNecromancer(plr)
    -- ability related constants
    local chillDamageMin = 9
    local chillDamageMax = 11
    local chillDuration = 90 -- 3 seconds
    local dfreezeRange = 200
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Necromancer"
    class.BaseHealth = 50
    class.AttackTimeout = 22
    class.MoveSpeedMul = 0.65
    class.AttackHandler = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            -- actual attack code
            local mx, my = unitPos(mob)
            local vx = x - mx
            local vy = y - my
            local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * 20
            vy = vy / vlen * 20
            if vx == vx and vy == vy then
                soundMake(907, mx, my)
                local shot = createObject("LightningBolt", mx, my)
                unitSetParent(shot, mob)
                unitSetLookAngle(shot, unitGetLookAngle(mob))
                setPtrFloat(shot, 0x50, vx)
                setPtrFloat(shot, 0x54, vy)
                unitAddRecentlyCreated()
            end
        end
    end
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(8, 1)
    class.AbilityNames[1] = "Chill Touch"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0x48)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            if (proj) then
                local projUC = getPtrPtr(proj, 0x2EC)
                local aim = utilFindClosestHostileMonster(x, y, mob, 200)
                if (aim) then setPtrPtr(projUC, 4, aim) end
                setPtrInt(projUC, 0xC, 0) -- reset spell type
                _G.unitFlyActivate[proj] = function()
                    -- actual ability code
                    aim = getPtrPtr(projUC, 4) -- spell may be inversed
                    if (aim) then
                        if (utilIsMagicVulnerable(aim)) then
                            buffApply(aim, 25, chillDuration) -- ENCHANT_FREEZE
                            unitDamageHack(aim, NOX_DAMAGE_TYPE.DRAIN, math.random(chillDamageMin, chillDamageMax))
                        end
                    end
                end
            end
        end
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(15, 1)
    class.AbilityNames[2] = "Deep Freeze"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            local mx, my = unitPos(mob)
            local applyFreeze = function(p)
                if (unitIsMonster(p) and unitCanAttack(mob, p)) then 
                    unitDamageHack(p, NOX_DAMAGE_TYPE.DRAIN, math.random(chillDamageMin, chillDamageMax))
                    buffApply(p, 25, chillDuration) -- ENCHANT_FREEZE
                    netPointFx(0x81, unitPos(p))
                end 
            end
            unitGetAround(applyFreeze, mx - dfreezeRange, my - dfreezeRange, mx + dfreezeRange, my + dfreezeRange)
            netPointFx(0xA3, mx, my)
            soundMake(913, mx, my)
        end
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnImp(plr)
    -- ability related constants
    local cloneLifetime = 300
    local numOfClones = 5
    local activeClones = {}
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Imp"
    class.BaseHealth = 25
    class.AttackTimeout = 28
    class.MoveSpeedMul = 0.75
    class.AttackHandler = function(mob, x, y)
        monsterDefaultMissileAttack(mob, x, y)
        -- let clones also shoot
        for i = 1, #activeClones do
            if not unitIsDead(activeClones[i]) then
                monsterDefaultMissileAttack(activeClones[i], x, y)
            end
        end
    end
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(20, 1)
    class.AbilityNames[1] = "Clone"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        netPointFx(0x96, unitPos(mob))
        local cHP = unitGetHPSafe(mob) / 2
        if (cHP < 1) then cHP = 1 end
        activeClones = {} -- TODO table.remove
        
        for n = 1, numOfClones do
            local imp = createObject("Imp", unitPos(mob))
            unitSetMaxHealth(imp, cHP)
            if (plr) then
                unitBecomePet(imp, plr)
            else
                unitSetParent(imp, mob)
            end
            unitSetAgression(imp, 0)
            table.insert(activeClones, imp)
            unitOnDie(imp, function(diedImp, attacker) netPointFx(0x96, unitPos(diedImp)) unitDecay(diedImp, 15) end)
            setTimeout(function() if not unitIsDead(imp) then unitDamagePure(imp, 999) end end, cloneLifetime)
        end
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(2, 1)
    class.AbilityNames[2] = "Swap Location"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        if (#activeClones > 0) then
            local mx, my, cx, cy, rng
            rng = activeClones[math.random(1, #activeClones)]
            if (objectExists(rng) and objectExists(mob)) then
                mx, my = unitPos(mob)
                cx, cy = unitPos(rng)
                unitMove(mob, cx, cy)
                unitMove(rng, mx, my)
                return false
            end
        end
        return true
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnGargoyle(plr)
    -- ability related constants
    local vampiricArrowHealthCoeff = 0.35
    local vampiricArrowBonusDamage = 10
    local chargedArrowStunDur = 80
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "EvilCherub"
    class.BaseHealth = 40
    class.AttackTimeout = 25
    class.AttackHandler = monsterDefaultMissileAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(4, 1)
    class.AbilityNames[1] = "Vampiric Arrow"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local shot = createObject("CherubArrow", unitPos(mob))
        unitSetParent(shot, mob)
        unitSetLookAngle(shot, unitGetLookAngle(mob))
        unitOnCollide(shot, function(shot2, targ)
            local owner = unitGetParent(shot2)
            local collData = getPtrPtr(shot2, 0x2BC)
            local dmg = getPtrInt(collData, 0)
            unitDelete(shot2)
            -- validation block
            if (not objectExists(owner)) then return true end
            if (not objectExists(targ)) then return true end
            -- end validation
            if (unitIsMonster(targ) and not unitTestBuff(targ, 23)) then
                unitDamageHack(targ, 13, dmg, owner)
                unitAddHP(owner, dmg) -- 100% vampirism
                local ox, oy = unitPos(owner)
                for i = 1, 3 do netRayFx(0x91, ox, oy, unitPos(targ)) end
                soundMake(5, unitPos(targ))
            end
        end)
        local mx, my = unitPos(mob)
        local vx, vy
        vx = x - mx
        vy = y - my
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * 50
        vy = vy / vlen * 50
        if vx == vx and vy == vy then
            setPtrFloat(shot, 0x50, vx)
            setPtrFloat(shot, 0x54, vy)
            unitAddRecentlyCreated() 
            setTimeout(function() if objectExists(shot) then unitDelete(shot) end end, 30)
        else
            unitDelete(shot)
        end -- test for -1.#IND
        local dmg = math.ceil(unitGetHPSafe(mob) * vampiricArrowHealthCoeff)
        unitDamagePure(mob, dmg)
        dmg = dmg + vampiricArrowBonusDamage
        arrowSetDamage(shot, dmg)
        soundMake(575, mx, my)
        soundMake(896, mx, my)
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(6, 1)
    class.AbilityNames[2] = "Charged Arrow"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local shot = createObject("CherubArrow", unitPos(mob))
        unitSetParent(shot, mob)
        unitSetLookAngle(shot, unitGetLookAngle(mob))
        unitOnCollide(shot, function(shot2, targ)
            local owner = unitGetParent(shot2)
            unitDelete(shot2)
            -- validation block
            if (not objectExists(owner)) then return true end
            if (not objectExists(targ)) then return true end
            -- end validation
            if (unitIsMonster(targ) and not unitTestBuff(targ, 23)) then
                if (not unitTestBuff(targ, 19)) then
                    buffApply(targ, 5, chargedArrowStunDur)
                    unitDamageHack(targ, NOX_DAMAGE_TYPE.IMPACT, 10, owner)
                end
                local px = getPtrFloat(shot, 0x50) / 4
                local py = getPtrFloat(shot, 0x54) / 4
                setPtrFloat(targ, 0x50, px)
                setPtrFloat(targ, 0x54, py)
            end
        end)
        local mx, my = unitPos(mob)
        local vx, vy
        vx = x - mx
        vy = y - my
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * 55
        vy = vy / vlen * 55
        if vx == vx and vy == vy then
            setPtrFloat(shot, 0x50, vx)
            setPtrFloat(shot, 0x54, vy)
            unitAddRecentlyCreated() 
            setTimeout(function() if objectExists(shot) then unitDelete(shot) end end, 30)
        else
            unitDelete(shot)
        end -- test for -1.#IND
        soundMake(886, unitPos(mob))
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnDryad(plr)
    -- ability related constants
    local lightningLength = 100
    local lightningMaxBounce = 4
    local naturesGiftDuration = 150 -- 5 sec
    local stunDuration = 60
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "WizardGreen"
    class.BaseHealth = 40
    class.AttackTimeout = 30
    class.AttackHandler = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            -- actual attack code
            local mx, my = unitPos(mob)
            soundMake(220, mx, my)
            local vx = x - mx
            local vy = y - my
            local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * 18
            vy = vy / vlen * 18
            if vx == vx and vy == vy then
                local shot = createObject("YellowStarShot", mx, my)
                setPtrInt(getPtrPtr(shot, 0x2BC), 0, 10) -- damage
                unitSetParent(shot, mob)
                unitSetLookAngle(shot, unitGetLookAngle(mob))
                setPtrFloat(shot, 0x50, vx)
                setPtrFloat(shot, 0x54, vy)
                unitAddRecentlyCreated()
            end
        end
    end
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(11, 1)
    class.AbilityNames[1] = "Green Lightning"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local mx, my = unitPos(mob)
        local vx = utilCosFromDirect(unitGetLookAngle(mob))
        local vy = utilSinFromDirect(unitGetLookAngle(mob))
        local tx = mx + vx * lightningLength
        local ty = my + vy * lightningLength
        if not mapTraceRay(mx, my, tx, ty, 5) then soundMake(231, mx, my) return true end -- fail
        unitActionPop(mob) utilSetActionCastTo(mob, tx, ty, 0) -- animation
        local damage = 23
        local secondary = {} -- secondary targets
        local secTargs = 0
        local distance = lightningLength
        local fn = function(p)
            if (unitIsMonster(p) and unitCanAttack(mob, p)) then
                secTargs = secTargs + 1
                if (secTargs <= lightningMaxBounce) then secondary[secTargs] = p end
            end
        end -- get nearby creatures, up to limit
        _G.unitOnCastList[mob] = function()
            netGreenLightning(mx, my, tx, ty, 5)
            soundMake(221, mx, my)
            unitGetAround(fn, tx - distance, ty - distance, tx + distance, ty + distance)
            local sec
            for i = 1, secTargs do
                sec = secondary[i]
                if (sec) then
                    mx = tx
                    my = ty
                    local tx2, ty2 = unitPos(sec)
                    if mapTraceRay(mx, my, tx2, ty2, 5) then
                        tx = tx2
                        ty = ty2
                        netGreenLightning(mx, my, tx, ty, 5)
                        if utilIsMagicVulnerable(sec) then
                            unitDamageHack(sec, NOX_DAMAGE_TYPE.ELECTRIC, damage, mob)
                        end
                    end
                end
            end
        end
        return false
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(7, 1)
    class.AbilityNames[2] = "Frog"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0x48)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            if (proj) then
                local projUC = getPtrPtr(proj, 0x2EC)
                local aim = utilFindClosestHostileMonster(x, y, mob, 200)
                if (aim) then setPtrPtr(projUC, 4, aim) end
                setPtrInt(projUC, 0xC, 0) -- reset spell type
                _G.unitFlyActivate[proj] = function()
                    -- actual ability code
                    aim = getPtrPtr(projUC, 4) -- spell may be inversed
                    if (aim) then
                        if (utilIsMagicVulnerable(aim)) then
                            netPointFx(0x8A, unitPos(aim))
                            soundMake(158, unitPos(aim))
                            utilApplyCustomEffect(aim, CustomEffectID.MORPH_FROG, stunDuration)
                        end
                    end
                end
            end
        end
    end
    class.AbilityUseTimeout[3] = utilSecToFrames(11, 1)
    class.AbilityNames[3] = "Nature's Gift"
    class.AbilityKeyHandlers[3] = function(mob, x, y)
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 41)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            if (proj) then
                local projUC = getPtrPtr(proj, 0x2EC)
                local aim = utilFindClosestFriendlyMonster(x, y, mob, 200)
                if (aim) then setPtrPtr(projUC, 4, aim) end
                setPtrInt(projUC, 0xC, 0) -- reset spell type
                _G.unitFlyActivate[proj] = function()
                    -- actual ability code
                    aim = getPtrPtr(projUC, 4)
                    if (aim) then
                        if (MobArena.MonsterToPlayerMap[aim]) then
                            utilApplyCustomEffect(aim, CustomEffectID.NATURES_GIFT, naturesGiftDuration)
                            soundMake(123, unitPos(aim))
                        end
                    end
                end
            end
        end
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnSpider(plr)
    -- ability related constants
    local arachnaRange = 120
    local arachnaMaxSpiders = 5
    local arachnaLifetime = 200
    local arachnaSpiderHealth = 10
    local venomSpitDamage = 10
    local venomSpitFlightSpeed = 9.5
    local venomSpitSlowDuration = 50
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Spider"
    class.BaseHealth = 70
    class.AttackTimeout = 20
    class.MoveSpeedMul = 0.7
    class.AttackHandler = monsterDefaultMeleeAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(15, 1)
    class.AbilityNames[1] = "Arachnaphobia"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local targets = 0
        local mx, my = unitPos(mob)
        local fn = function(p)
            if (unitIsMonster(p) and unitCanAttack(mob, p)) then
                targets = targets + 1
                if (targets <= arachnaMaxSpiders) then
                    local px, py = unitPos(p)
                    if not mapTraceRay(mx, my, px, py, 5) then return end
                    -- spawn a spider
                    local spi = createObject("SmallAlbinoSpider", px, py)
                    setPtrFloat(spi, 0x50, math.random(-10, 10))
                    setPtrFloat(spi, 0x54, math.random(-10, 10))
                    unitSetMaxHealth(spi, arachnaSpiderHealth)
                    unitSetAgression(spi, 0)
                    unitSetParent(spi, mob)
                    unitAddRecentlyCreated()
                    setPtrPtr(getPtrPtr(spi, 0x2EC), 0x4AC, p) -- set target
                    utilSetActionFight(spi, p)
                    unitOnDie(spi, function(died, attacker) unitDecay(died, 15) end)
                    setTimeout(function() if not unitIsDead(spi) then unitDamagePure(spi, 999) end end, arachnaLifetime)
                end
            end
        end
        unitGetAround(fn, mx - arachnaRange, my - arachnaRange, mx + arachnaRange, my + arachnaRange)
        if (targets == 0) then soundMake(231, mx, my) return true end
    end
    class.VenomSpitCollide = function(spit, coll)
        local owner = unitGetParent(spit)
        if (not objectExists(owner)) then return true end
        if coll ~= owner then soundMake(351, unitPos(spit)) unitDelete(spit) end
        if (not objectExists(coll)) then return true end
        
        if (unitIsMonster(coll) and unitCanAttack(owner, coll)) then
            unitDamageHack(coll, NOX_DAMAGE_TYPE.POISON, venomSpitDamage, owner)
            buffApply(coll, 4, venomSpitSlowDuration)
        end
        
        return true
    end
    class.VenomSpitUpdate = function(spit)
        if (not objectExists(spit)) then return end
        local timeCreated = getPtrInt(spit, 0x80)
        if getFrameCounter() - timeCreated > 40 then
            unitDelete(spit)
            return
        end
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(6, 1)
    class.AbilityNames[2] = "Venomous Spit"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local mx, my = unitPos(mob)
        soundMake(0x15C, mx, my)
        local vx = x - mx
        local vy = y - my
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * venomSpitFlightSpeed
        vy = vy / vlen * venomSpitFlightSpeed
        local proj = createObject("SpiderSpit", mx, my)
        unitSetParent(proj, mob)
        setPtrFloat(proj, 0x50, vx)
        setPtrFloat(proj, 0x54, vy)
        unitOnUpdateHack(proj, class.VenomSpitUpdate)
        unitOnCollide(proj, class.VenomSpitCollide)
        unitAddRecentlyCreated()
    end

    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnArchmage(plr)
    -- ability related constants
    local restorationRange = 160
    local restorationHeal = 20
    local restorationInvulnDur = 30
    local flameDuration = 60
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "WizardWhite"
    class.BaseHealth = 40
    class.AttackTimeout = 22
    class.MoveSpeedMul = 0.7
    class.AttackHandler = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            -- actual attack code
            local mx, my = unitPos(mob)
            soundMake(218, mx, my)
            local vx = x - mx
            local vy = y - my
            local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * 20
            vy = vy / vlen * 20
            if vx == vx and vy == vy then
                local shot = createObject("PitifulFireball", mx, my)
                unitSetParent(shot, mob)
                unitSetLookAngle(shot, unitGetLookAngle(mob))
                setPtrFloat(shot, 0x50, vx)
                setPtrFloat(shot, 0x54, vy)
                unitAddRecentlyCreated()
            end
        end
    end
    -- special
    class.ImmolationActive = false
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(10, 1)
    class.AbilityNames[1] = "Force Field"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 41)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            if (proj) then
                local projUC = getPtrPtr(proj, 0x2EC)
                local aim = utilFindClosestFriendlyMonster(x, y, mob, 200)
                if (aim) then setPtrPtr(projUC, 4, aim) end
                setPtrInt(projUC, 0xC, 0) -- reset spell type
                _G.unitFlyActivate[proj] = function()
                    -- actual ability code
                    aim = getPtrPtr(projUC, 4)
                    if (aim) then
                        spellApply(aim, 51, aim) -- SPELL_SHIELD
                    end
                end
            end
        end
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(12, 1)
    class.AbilityNames[2] = "Restoration"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            local mx, my = unitPos(mob)
            netPointFx(0xA0, mx, my)
            soundMake(773, mx, my)
            local fn = function(p) 
                if unitIsMonster(p) and (not unitCanAttack(mob, p)) then
                    unitAddHP(p, restorationHeal)
                    soundMake(754, unitPos(p))
                    if (not unitTestBuff(p, 23)) then
                        buffApply(p, 23, restorationInvulnDur) -- ENCHANT_INVULNERABLE
                    end
                end
            end
            unitGetAround(fn, mx - restorationRange, my - restorationRange, mx + restorationRange, my + restorationRange)
        end
    end
    class.AbilityUseTimeout[3] = utilSecToFrames(11, 1)
    class.AbilityNames[3] = "Immolation"
    class.AbilityKeyHandlers[3] = function(mob, x, y)
        local mx, my = unitPos(mob)
        if not class.ImmolationActive then
            if mapTraceRay(mx, my, x, y, 5) then
                soundMake(9, mx, my)
                class.ImmolationActive = true
                class.ImmolationCastTime = getFrameCounter()
                class.ImmolationFlame = createObject("Flame", x, y)
                unitToggleFlag(class.ImmolationFlame, NOX_UNIT_FLAGS.SHORT) -- damage flying units
                unitSetParent(class.ImmolationFlame, mob)
                unitDecay(class.ImmolationFlame, flameDuration)
                unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0) -- animation
                utilApplyCustomEffect(mob, CustomEffectID.CHANNELING, flameDuration)
            else
                -- wrong target
                soundMake(231, mx, my)
                return true
            end
        end
        
        --[[unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        local mx, my = unitPos(mob)
        if not mapTraceRay(mx, my, x, y, 5) then soundMake(231, mx, my) return true end
        _G.unitOnCastList[mob] = function()
            buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
            soundMake(9, x, y)
            local angle, fire, vx, vy
            for angle = 0, 256, 10 do
                -- calculate initial position
                vx = utilCosFromDirect(angle) * fireStartRadius + x
                vy = utilSinFromDirect(angle) * fireStartRadius + y
                if (mapTraceRay(mx, my, vx, vy, 5)) then
                    -- spawn a flame
                    fire = createObject(fireNames[math.random(1, #fireNames)], vx, vy)
                    -- set parent
                    unitSetParent(fire, mob)
                    -- set movement vector
                    vx = utilCosFromDirect(angle) * fireExpandSpeed
                    vy = utilSinFromDirect(angle) * fireExpandSpeed
                    setPtrFloat(fire, 0x50, vx)
                    setPtrFloat(fire, 0x54, vy)
                    setPtrFloat(fire, 0x70, 0)
                    -- activate linear movement prediction
                    setPtrInt(fire, 8, bitOr(getPtrInt(fire, 8), 0x40000000))
                    ptrCall2(0x0523530, fire, fire)
                    unitDecay(fire, fireLifetime)
                end
            end
        end]]--
    end
    class.EveryFrame = function()
        if class.ImmolationActive then
            if objectExists(class.ImmolationFlame) then
                if not utilCheckCustomEffect(class.MonsterPtr, CustomEffectID.CHANNELING) then
                    unitDecay(class.ImmolationFlame, 1)
                    class.ImmolationActive = false
                    return
                end
                local mx, my = unitPos(class.ImmolationFlame)
                netRayFx(0x91, mx, my, unitPos(class.MonsterPtr))
                local plr = MobArena.MonsterToPlayerMap[class.MonsterPtr]
                if plr then
                    local x, y = playerMouse(plr)
                    if (mapTraceRay(mx, my, x, y, 1) and distance(mx, my, x, y) < 200) then
                        unitMove(class.ImmolationFlame, x, y)
                        unitSetLookAngle(class.MonsterPtr, directGet(x, y, unitPos(class.MonsterPtr)))
                    end
                end
            else
                class.ImmolationActive = false
            end
        end
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnShade(plr)
    -- ability related constants
    local jumpSearchRange = 200
    local jumpTargSlowDuration = 90
    local shadowWalkDuration = 140
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Shade"
    class.BaseHealth = 60
    class.AttackTimeout = 20
    class.MoveSpeedMul = 0.67
    class.AttackHandler = monsterDefaultMeleeAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(10, 1)
    class.AbilityNames[1] = "Jump"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local mx, my = unitPos(mob)
        local hostile = utilFindClosestHostileMonster(mx, my, mob, jumpSearchRange)
        if (hostile) then
            -- hop effect
            unitRaiseHack(mob, 1)
            setPtrFloat(mob, 0x6C, 5)
            -- calculate jump vector
            local hx, hy = unitPos(hostile)
            local vx, vy
            vx = hx - mx
            vy = hy - my
            vx = vx / 2
            vy = vy / 2
            -- ignore obstacles
            unitToggleFlag(mob, NOX_UNIT_FLAGS.NO_COLLIDE)
            unitToggleFlag(mob, NOX_UNIT_FLAGS.AIRBORNE)
            setTimeout(function() if objectExists(mob) then
                unitToggleFlag(mob, NOX_UNIT_FLAGS.NO_COLLIDE)
                unitToggleFlag(mob, NOX_UNIT_FLAGS.AIRBORNE)
            end end, 6) -- remove flag after 6 frames
            buffApply(hostile, 4, jumpTargSlowDuration)
            soundMake(772, hx, hy)
            if vx == vx and vy == vy then
                setPtrFloat(mob, 0x50, vx)
                setPtrFloat(mob, 0x54, vy)
            end
            return false
        end
        soundMake(231, mx, my)
        return true
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(22, 1)
    class.AbilityNames[2] = "Shadow Walk"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffApply(mob, 23, shadowWalkDuration) -- ENCHANT_INVULNERABLE
        buffApply(mob, 29, shadowWalkDuration) -- ENCHANT_ANTI_MAGIC
        unitToggleFlag(mob, NOX_UNIT_FLAGS.OWNER_VISIBLE)
        class.ShadowWalkActive = true
        setTimeout(function() 
            class.ShadowWalkActive = false
            if objectExists(mob) then unitToggleFlag(mob, NOX_UNIT_FLAGS.OWNER_VISIBLE) end
            end, shadowWalkDuration)
    end
    class.EveryFrame = function()
        if class.ShadowWalkActive and not unitIsDead(class.MonsterPtr) then
            -- regenerate 1 hp every 3 frames
            if math.mod(getFrameCounter(), 3) == 0 then
                unitAddHP(class.MonsterPtr, 1)
            end
        end
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnWasp(plr)
    -- ability related constants
    local waspNestHP = 50
    local waspSpawnDelay = 100
    local waspSpawnLimit = 3
    local waspMinionHP = 12
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Wasp"
    class.BaseHealth = 25
    class.AttackTimeout = 12
    class.MoveSpeedMul = 0.7
    class.AttackHandler = monsterDefaultMeleeAttack
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(31, 1)
    class.AbilityNames[1] = "Build Nest"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local mx, my = unitPos(mob)
        if class.WaspNest and objectExists(class.WaspNest) then unitDamagePure(class.WaspNest, 999) end
        class.WaspNest = createObject("WaspNest", mx, my)
        class.WaspLastSpawn = getFrameCounter() + waspSpawnDelay
        unitToggleFlag(class.WaspNest, NOX_UNIT_FLAGS.SHORT)
        unitHP(class.WaspNest, waspNestHP)
        unitSetParent(class.WaspNest, mob)
        netPointFx(0x8A, mx, my)
        soundMake(835, mx, my)
    end
    class.WaspWorkerDieHandler = function(killed, killer)
        class.WaspsActive = class.WaspsActive - 1
    end
    class.EveryFrame = function()
        if not class.WaspNest then return end
        if unitIsDead(class.WaspNest) then return end
        if getFrameCounter() > class.WaspLastSpawn then
            if class.WaspsActive < waspSpawnLimit then
                -- spawn new wasp follower
                class.WaspLastSpawn = getFrameCounter() + waspSpawnDelay
                class.WaspsActive = class.WaspsActive + 1
                local w = createObject("Wasp", unitPos(class.WaspNest))
                unitSetMaxHealth(w, waspMinionHP)
                if (plr) then
                    unitBecomePet(w, plr)
                else
                    unitSetParent(w, class.MonsterPtr)
                end
                unitAddRecentlyCreated()
                unitSetAgression(w, 0.83)
                unitSetFollow(w, class.MonsterPtr)
                unitOnDie(w, class.WaspWorkerDieHandler)
            end
        end
    end
    class.MonsterDie = function()
        if class.WaspNest and objectExists(class.WaspNest) then unitDamagePure(class.WaspNest, 999) end
    end
    -- special
    class.WaspsActive = 0
    class.WaspLastSpawn = 0
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnWizard(plr)
    -- ability related constants
    local blinkMaxDistance = 200
    local strangleDuration = 30
    local strangleMaxDistance = 250
    local boulderStartDist = 44
    local boulderMoveSpeed = 8
    local boulderLifetime = 55
    local boulderAirTime = 6
    local boulderDmgPerTick = 1
    local boulderLandArea = 70
    local boulderLandDamage = 16
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Wizard"
    class.BaseHealth = 50
    class.AttackTimeout = 18
    class.MoveSpeedMul = 1
    class.AttackHandler = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            -- actual attack code
            local mx, my = unitPos(mob)
            soundMake(218, mx, my)
            local vx = x - mx
            local vy = y - my
            local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * 20
            vy = vy / vlen * 20
            if vx == vx and vy == vy then
                local shot = createObject("PitifulFireball", mx, my)
                unitSetParent(shot, mob)
                unitSetLookAngle(shot, unitGetLookAngle(mob))
                setPtrFloat(shot, 0x50, vx)
                setPtrFloat(shot, 0x54, vy)
                unitAddRecentlyCreated()
            end
        end
    end
    
    -- ability descriptors
    class.AbilityUseTimeout[1] = utilSecToFrames(14, 1)
    class.AbilityNames[1] = "Blink"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        local mx, my = unitPos(mob)
        -- check visibility and distance
        if not mapTraceRay(mx, my, x, y, 9) then soundMake(231, mx, my) return true end
        if distance(mx, my, x, y) > blinkMaxDistance then
            local vx = x - mx
            local vy = y - my
            local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * blinkMaxDistance + mx
            vy = vy / vlen * blinkMaxDistance + my
            x = vx
            y = vy
        end
        -- cast
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            netPointFx(0x89, mx, my)
            netPointFx(0x89, x, y)
            soundMake(0x98, mx, my)
            unitMove(mob, x, y)
        end
    end
    class.StrangleTarget = nil
    class.StrangleEffect = false
    class.StrangleUpdate = function(obj)
        if (not objectExists(obj)) then return end
        local owner = unitGetParent(obj)
        if unitIsDead(owner) then unitDelete(obj) return end
        -- check time
        local timeCreated = getPtrInt(obj, 0x80)
        if getFrameCounter() - timeCreated > strangleDuration or unitIsDead(class.StrangleTarget) then
            unitDelete(obj)
            return
        end
        local mx, my = unitPos(owner)
        local tx, ty = unitPos(class.StrangleTarget)
        -- check distance
        if distance(mx, my, tx, ty) > strangleMaxDistance then
            unitDelete(obj)
            soundMake(231, mx, my)
            return
        end
        -- visual effects
        if monsterGetCurrentAction(owner) ~= 19 then
            -- animation
            unitActionPop(owner) utilSetActionCastTo(owner, tx, ty, 0)
            -- fix spell duplication
            _G.unitOnCastList[owner] = nil
        end
        if math.mod(getFrameCounter(), 10) ~= 0 then
            netRayFx(0x8C, mx, my, tx, ty) -- fx
            soundMake(0x62, tx, ty) -- audio
        end
        --if not class.StrangleEffect then netStartDurationSpell(owner, 0x18, class.StrangleTarget, 3) class.StrangleEffect = true end
        -- deal damage
        unitDamageHack(class.StrangleTarget, NOX_DAMAGE_TYPE.DRAIN, 1, nil)
        setPtrPtr(class.StrangleTarget, 0x208, owner) -- damage parent
        if not unitTestBuff(class.StrangleTarget, 5) then
            -- apply stun
            buffApply(class.StrangleTarget, 5, 2) -- ENCHANT_HELD
        else
            -- or increase duration
            setPtrInt(class.StrangleTarget, 0x158 + 10, 2)
        end
    end
    class.AbilityUseTimeout[2] = utilSecToFrames(12, 1)
    class.AbilityNames[2] = "Strangle"
    class.AbilityKeyHandlers[2] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0x48)
        _G.unitOnCastList[mob] = function(spellID, caster, target, proj, tx, ty)
            if (proj) then
                local projUC = getPtrPtr(proj, 0x2EC)
                local aim = utilFindClosestHostileMonster(x, y, mob, 200)
                if (aim) then setPtrPtr(projUC, 4, aim) end
                setPtrInt(projUC, 0xC, 0) -- reset spell type
                _G.unitFlyActivate[proj] = function()
                    local parent = unitGetParent(proj)
                    if unitIsDead(parent) then return end -- verify
                    aim = getPtrPtr(projUC, 4) -- spell may be inversed
                    if (aim) then
                        -- actual ability code
                        if (utilIsMagicVulnerable(aim)) then
                            local dummy = createObject("ImaginaryCaster", x, y)
                            unitSetParent(dummy, mob)
                            class.StrangleTarget = aim
                            class.StrangleEffect = false
                            utilApplyCustomEffect(mob, CustomEffectID.CHANNELING, strangleDuration)
                            unitOnUpdateHack(dummy, class.StrangleUpdate)
                            unitAddRecentlyCreated()
                        end
                    end
                end
            end
        end
    end
    class.BoulderCollideHandler = function(boulder, coll)
        local owner = unitGetParent(boulder)
        if not (objectExists(owner) and objectExists(coll)) then return true end
        
        -- check if boulder is not landed yet
        if getFrameCounter() - getPtrInt(boulder, 0x80) < boulderAirTime then return end
        
        if unitIsMonster(coll) and owner ~= coll then
            unitDamageHack(coll, NOX_DAMAGE_TYPE.CRUSH, boulderDmgPerTick, owner)
        end
    end
    class.AbilityUseTimeout[3] = utilSecToFrames(19, 1)
    class.AbilityNames[3] = "Summon Boulder"
    class.AbilityKeyHandlers[3] = function(mob, x, y)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        local mx, my = unitPos(mob)
        -- check visibility and distance
        local vx = x - mx
        local vy = y - my
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * boulderStartDist + mx
        vy = vy / vlen * boulderStartDist + my
        local bx = vx
        local by = vy
        if not mapTraceRay(mx, my, bx, by, 9) then soundMake(231, mx, my) return true end
        -- cast
        unitActionPop(mob) utilSetActionCastTo(mob, x, y, 0)
        _G.unitOnCastList[mob] = function()
            vx = x - mx
            vy = y - my
            vlen = math.sqrt(vx ^ 2 + vy ^ 2)
            vx = vx / vlen * boulderMoveSpeed
            vy = vy / vlen * boulderMoveSpeed
            if vx == vx and vy == vy then
                local boulder = createObject("BoulderIndestructible", bx, by)
                unitRaiseHack(boulder, 120) -- initial height
                setPtrFloat(boulder, 0x6C, -7) -- fall speed
                unitSetParent(boulder, mob)
                setPtrFloat(boulder, 0x50, vx) -- moving vector
                setPtrFloat(boulder, 0x54, vy)
                unitAddRecentlyCreated()
                local boulderFn = function(p)
                    if (unitIsMonster(p) and unitCanAttack(mob, p)) then 
                        unitSplashDamage(p,
                            bx, by, 
                            boulderLandArea, 
                            boulderLandArea / 2, 
                            boulderLandDamage, 
                            NOX_DAMAGE_TYPE.CRUSH,
                            mob)
                    end
                end
                setTimeout(function() 
                    bx, by = unitPos(boulder)
                    soundMake(0x30, bx, by)
                    netPointFx(0x8A, bx, by)
                    netSendEarthquake(bx, by, 25)
                    if not unitIsDead(mob) then
                        -- apply land damage
                        unitGetAround(boulderFn, bx - boulderLandArea, by - boulderLandArea, bx + boulderLandArea, by + boulderLandArea)
                    end
                end, boulderAirTime)
                setTimeout(function()
                    if objectExists(boulder) then
                        bx, by = unitPos(boulder)
                        soundMake(757, bx, by)
                        netPointFx(0x8A, bx, by)
                        netSendEarthquake(bx, by, 17)
                        unitDelete(boulder)
                    end
                end, boulderLifetime)
                unitOnCollide(boulder, class.BoulderCollideHandler)
            end
        end
    end
    class.SpacePressed = function()
        local mx, my = unitPos(class.MonsterPtr)
        soundMake(0x1FF, mx, my)
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end

function MobArena.spawnFireKnight(plr)
    -- ability related constants
    local harpoonFlightSpeed = 18
    local harpoonPullSpeed = 22
    local harpoonFinishRange = 35
    local harpoonProjLifetime = 45
    local harpoonTrackRange = 20
    local harpoonFollowRange = 30
    local harpoonDamageInterval = 2
    
    -- set monster data
    local class = NewMonsterClass()
    class.ThingID = "Horrendous"
    class.BaseHealth = 65
    class.AttackTimeout = 35
    class.MoveSpeedMul = 0.5
    class.AttackHandler = monsterDefaultMeleeAttack
    
    -- ability descriptors
    class.HarpoonLockTarget = nil
    class.HarpoonReeled = false -- harpoon stuck in enemy
    class.HarpoonUpdate = function(harp)
        if (not objectExists(harp)) then return end
        local timeCreated = getPtrInt(harp, 0x80)
        if getFrameCounter() - timeCreated > harpoonProjLifetime then
            -- prevent this if harpoon is stuck
            if not class.HarpoonReeled then
                soundMake(998, unitPos(harp)) -- break sound
                unitDelete(harp)
                return
            end
        end
        local owner = unitGetParent(harp)
        if (not objectExists(owner)) then
            if class.HarpoonLockTarget and class.HarpoonReeled then
                -- ground the target
                if objectExists(class.HarpoonLockTarget) then
                    unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.NO_COLLIDE)
                    unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.AIRBORNE)
                    class.HarpoonReeled = false
                end
            end
            unitDelete(harp)
            return 
        end
        local hx, hy = unitPos(harp)
        -- search for possible targets
        if not class.HarpoonLockTarget then
            local fn = function(p) 
                if unitIsMonster(p) and unitCanSeeUnit(harp, p) and unitCanAttack(owner, p) then 
                    class.HarpoonLockTarget = p 
                end
            end
            unitGetAround(fn, hx - harpoonTrackRange, hy - harpoonTrackRange, hx + harpoonTrackRange, hy + harpoonTrackRange)
        else
            -- update if target found
            if not unitIsDead(class.HarpoonLockTarget) then
                local cx, cy = unitPos(class.HarpoonLockTarget)
                if class.HarpoonReeled then
                    local ox, oy = unitPos(owner)
                    -- check distance and then finish
                    if distance(ox, oy, cx, cy) < harpoonFinishRange or not mapTraceRay(ox, oy, cx, cy, 5) then
                        -- ground the target
                        unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.NO_COLLIDE)
                        unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.AIRBORNE)
                        soundMake(998, unitPos(harp)) 
                        unitDelete(harp)
                        return
                    end
                    -- flight effect
                    unitRaiseHack(class.HarpoonLockTarget, 25) -- initial height
                    setPtrFloat(class.HarpoonLockTarget, 0x6C, -5) -- fall speed
                    -- carry the victim to owner
                    local vx = ox - cx
                    local vy = oy - cy
                    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
                    vx = vx / vlen * harpoonPullSpeed
                    vy = vy / vlen * harpoonPullSpeed
                    if vx == vx and vy == vy then
                        -- move harpoon
                        unitMove(harp, cx + vx, cy + vy)
                        setPtrFloat(harp, 0x50, vx)
                        setPtrFloat(harp, 0x54, vy)
                        -- move victim
                        setPtrFloat(class.HarpoonLockTarget, 0x50, vx)
                        setPtrFloat(class.HarpoonLockTarget, 0x54, vy)
                    end
                    -- damage the victim
                    if math.mod(getFrameCounter(), harpoonDamageInterval) == 0 then
                        unitDamageHack(class.HarpoonLockTarget, NOX_DAMAGE_TYPE.IMPALE, 1, nil)
                        setPtrPtr(class.HarpoonLockTarget, 0x208, owner) -- damage parent
                    end
                else
                    -- if target is too far then reset
                    if distance(cx, cy, hx, hy) > harpoonFollowRange then class.HarpoonLockTarget = nil return end
                    -- follow target
                    local vx = getPtrFloat(harp, 0x50)
                    local vy = getPtrFloat(harp, 0x54)
                    local vx2 = cx - hx
                    local vy2 = cy - hy
                    local vlen = math.sqrt(vx2 ^ 2 + vy2 ^ 2)
                    vx2 = vx2 / vlen * harpoonFlightSpeed
                    vy2 = vy2 / vlen * harpoonFlightSpeed
                    if vx2 == vx2 and vy2 == vy2 then
                        -- smooth vector
                        vx = (vx + vx2) / 4
                        vy = (vy + vy2) / 4
                        setPtrFloat(harp, 0x50, vx)
                        setPtrFloat(harp, 0x54, vy)
                    end
                end
            else
                -- target lost - reset anything
                class.HarpoonLockTarget = nil
                class.HarpoonReeled = false
            end
        end
    end
    class.HarpoonCollide = function(harp, targ)
        local owner = unitGetParent(harp)
        -- validation block
        if (not objectExists(owner)) then return true end
        if (not objectExists(targ)) then
            if class.HarpoonLockTarget and class.HarpoonReeled then
                -- ground the target
                if objectExists(class.HarpoonLockTarget) then
                    unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.NO_COLLIDE)
                    unitToggleFlag(class.HarpoonLockTarget, NOX_UNIT_FLAGS.AIRBORNE)
                    class.HarpoonReeled = false
                end
            end
            soundMake(998, unitPos(harp))
            unitDelete(harp) 
            return true
        end
        -- end validation
        local mx, my = unitPos(owner)
        local tx, ty, vx, vy
        if (unitIsMonster(targ) and targ ~= owner) then
            if not class.HarpoonReeled then
                -- pull target onto air
                unitToggleFlag(targ, NOX_UNIT_FLAGS.NO_COLLIDE)
                unitToggleFlag(targ, NOX_UNIT_FLAGS.AIRBORNE)
                class.HarpoonLockTarget = targ
                class.HarpoonReeled = true
                soundMake(999, unitPos(targ))
            end
        end
        return true
    end
    class.AbilityUseTimeout[1] = utilSecToFrames(16, 1)
    class.AbilityNames[1] = "Harpoon"
    class.AbilityKeyHandlers[1] = function(mob, x, y)
        local mx, my = unitPos(mob)
        buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
        soundMake(177, unitPos(mob)) -- harpoon sound
        -- spawn projectile
        local harp = createObject("WeakArcherArrow", unitPos(mob))
        unitSetParent(harp, mob)
        class.HarpoonLockTarget = nil
        class.HarpoonReeled = false
        -- handlers
        unitOnUpdateHack(harp, class.HarpoonUpdate)
        unitOnCollide(harp, class.HarpoonCollide)
        -- velocity/facing
        local angle = directGet(x, y, unitPos(mob))
        unitSetLookAngle(mob, angle)
        unitSetLookAngle(harp, angle)
        local mx, my = unitPos(mob)
        local vx, vy
        vx = x - mx
        vy = y - my
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * harpoonFlightSpeed
        vy = vy / vlen * harpoonFlightSpeed
        if vx == vx and vy == vy then
            setPtrFloat(harp, 0x50, vx)
            setPtrFloat(harp, 0x54, vy)
            unitAddRecentlyCreated()
        else
            unitDelete(harp)
        end -- test for -1.#IND
    end
    
    -- spawn monster
    if (plr) then
        class.Respawn(unitGetTeam(plr))
        class.SetPlayer(plr)
        -- required for respawn to work
        MobArena.MonsterToSpawnClass[class.MonsterPtr] = class
    end
    return class
end