-- reward.lua
-- Random reward generator by KirConjurer (angrykirc)

local armorQuality = {'ArmorQuality1','ArmorQuality2','ArmorQuality3','ArmorQuality4','ArmorQuality5','ArmorQuality6'}
local powerColor = {'UserColor24','UserColor28','UserColor21','UserColor25','UserColor22','UserColor11'}
local leatherArmorNames = {'LeatherHelm','LeatherArmbands','LeatherArmor','LeatherLeggings','LeatherArmoredBoots'}
local plateArmorNames = {'SteelHelm','OrnateHelm','PlateArms','Breastplate','PlateLeggings','PlateBoots','ChainTunic','ChainLeggings'}
local armorEnchants = {'LightningProtect','FireProtect','Regeneration','Speed','PoisonProtect'}
local otherArmorNames = {'MedievalCloak','WizardRobe','WizardHelm','LeatherBoots','SteelShield','RoundShield'}
local weaponQuality = {'WeaponPower1','WeaponPower2','WeaponPower3','WeaponPower4','WeaponPower5','WeaponPower6'}
local weaponNames1 = {'Quiver','Longsword','Sword','StaffWooden','FanChakram','BattleAxe','MorningStar'}
local weaponNames2 = {'WarHammer','GreatSword','OgreAxe'}
local magicWandNames = {'LesserFireballWand','FireStormWand','ForceStormWand'}
local weaponEnchants = {'Stun','Fire','Lightning','Impact','Vampirism','Venom'} -- ManaSteal and Confuse excluded
local spellIdList = {1,5,10,16,22,24,27,29,31,35,36,37,38,39,41,43,50,51,53,54,55,63,64,65}
rewardGeneratorBaseRarity = 1

local function randomRarity(rarity, maximum)
 if (math.random(0,3) == 0) then
  rarity = rarity + 1
 else
  return rarity
 end
 if (rarity >= maximum) then
  return rarity
 else
  return randomRarity(rarity,maximum)
 end
end

local function randomArmorEnchant(maxLevel)
 local ench=""
 if (math.random(0,rewardGeneratorBaseRarity)==0) then return ench end
 ench = armorEnchants[math.random(1,#armorEnchants)]
 ench = ench .. randomRarity(1,maxLevel)
 return ench
end

local function randomWeaponEnchant(maxLevel)
 local ench=""
 if (math.random(0,rewardGeneratorBaseRarity)==0) then return ench end
 ench = weaponEnchants[math.random(1,#weaponEnchants)]
 ench = ench .. randomRarity(1,maxLevel)
 return ench
end

function clearInventory(unit)
    for item in getInv(unit) do
        unitDelete(item)
    end
end

function randomSpellBook(inv)
 local obj = createObjectIn("SpellBook", inv)
 local spell = spellIdList[math.random(1,#spellIdList)]
 local t = getPtrPtr(obj, 0x2E0)
 setPtrInt(t, 0, getPtrInt(t,0) + spell)
 return obj
end

function randomWeaponT2(inv)
 local name = weaponNames2[math.random(1,#weaponNames2)]
 local rarity = randomRarity(rewardGeneratorBaseRarity,6)
 local obj = createObjectIn(name,inv)
 
 local firstEnchant=randomWeaponEnchant(3)
 local secondEnchant=""
 if (firstEnchant~="") then secondEnchant=randomWeaponEnchant(3) end 
 itemEnchants(obj,weaponQuality[rarity],powerColor[rarity],firstEnchant,secondEnchant)
 return obj
end

function randomWeaponT1(inv)
 local name = weaponNames1[math.random(1,#weaponNames1)]
 local rarity = randomRarity(rewardGeneratorBaseRarity,6)
 local obj=createObjectIn(name,inv)
 
 local firstEnchant=randomWeaponEnchant(3)
 local secondEnchant=""
 if (firstEnchant~="") then secondEnchant=randomWeaponEnchant(3) end 
 itemEnchants(obj,weaponQuality[rarity],powerColor[rarity],firstEnchant,secondEnchant)
 return obj
end

function randomMagicItem(inv)
 local name
 if (math.random(0,2) == 0) then
  name = magicWandNames[math.random(1,#magicWandNames)]
 else
  name = "RedPotion"
 end
 return createObjectIn(name,inv)
end

function randomOtherArmor(inv)
 local name = otherArmorNames[math.random(1,#otherArmorNames)]
 local rarity = randomRarity(rewardGeneratorBaseRarity,6)
 local obj=createObjectIn(name,inv)
 
 local firstEnchant=randomArmorEnchant(3)
 local secondEnchant=""
 if (firstEnchant~="") then secondEnchant=randomArmorEnchant(4) end 
 itemEnchants(obj,armorQuality[rarity],powerColor[rarity],firstEnchant,secondEnchant)
 return obj
end

function randomPlateArmor(inv)
 local name = plateArmorNames[math.random(1,#plateArmorNames)]
 local rarity = randomRarity(rewardGeneratorBaseRarity,6)
 local obj=createObjectIn(name,inv)
 
 local firstEnchant=randomArmorEnchant(3)
 local secondEnchant=""
 if (firstEnchant~="") then secondEnchant=randomArmorEnchant(4) end 
 itemEnchants(obj,armorQuality[rarity],powerColor[rarity],firstEnchant,secondEnchant)
 return obj
end

function randomLeatherArmor(inv)
 local name = leatherArmorNames[math.random(1,#leatherArmorNames)]
 local rarity = randomRarity(rewardGeneratorBaseRarity,6)
 local obj=createObjectIn(name,inv)
 
 local firstEnchant=randomArmorEnchant(3)
 local secondEnchant=""
 if (firstEnchant~="") then secondEnchant=randomArmorEnchant(4) end 
 itemEnchants(obj,armorQuality[rarity],powerColor[rarity],firstEnchant,secondEnchant)
 return obj
end